#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import signal
import sys
import time

import concurrent.futures
import spur

import fuelpdsh
import fuelpdsh.ssh
import fuelpdsh.stream
import fuelpdsh.utils


LOG = fuelpdsh.logger(__name__)
"""Logger."""

GENTLE_STOP_TIMEOUT = 5
"""How long to wait before sending SIGTERM to SSH process."""


def command(hostnames, options, stop_ev):
    concurrency = options.concurrency
    if not concurrency:
        concurrency = len(hostnames)

    LOG.info("Execute with %d threads", concurrency)

    stdout_stream = fuelpdsh.stream.Stream(sys.stdout)
    stderr_stream = fuelpdsh.stream.Stream(sys.stderr)

    stdout_stream.run()
    stderr_stream.run()

    futures = []
    try:
        with concurrent.futures.ThreadPoolExecutor(concurrency) as pool:

            hostname_padding = max(len(host) for host in hostnames)
            for host in hostnames:
                stdout = stdout_stream.make_host_stream(host, hostname_padding)
                stderr = stderr_stream.make_host_stream(host, hostname_padding)
                future = pool.submit(run_on_host, host, options, stdout, stderr, stop_ev)
                futures.append(future)

            fuelpdsh.utils.wait_for_futures(futures, stop_ev)
    finally:
        stdout_stream.stop()
        stderr_stream.stop()
        for stream in stdout_stream, stderr_stream:
            stream.wait()

        return fuelpdsh.utils.futures_exit_code(futures)


def run_on_host(host, options, stdout, stderr, stop_ev):
    if stop_ev.is_set():
        return os.EX_OK

    str_command = " ".join(options.command)

    with fuelpdsh.ssh.get_ssh(host) as ssh:
        LOG.debug("Execute %s on host %s", str_command, host)

        try:
            process = ssh.spawn(options.command,
                                stdout=stdout, stderr=stderr, store_pid=True, allow_error=False)
            while process.is_running():
                if stop_ev.is_set():
                    raise KeyboardInterrupt
                time.sleep(0.5)
        except spur.NoSuchCommandError:
            LOG.error("No such command on host %s", host)
            raise
        except spur.RunProcessError as exc:
            stderr.write("\n{0}".format(exc.stderr_output))
            raise
        except KeyboardInterrupt:
            stop_ssh_process(host, process)
            raise
        except:
            LOG.exception("Problem with executing %s on host %s", str_command, host)
            raise
        else:
            return process.wait_for_result().return_code
        finally:
            stdout.flush()
            stderr.flush()


def stop_ssh_process(host, process):
    LOG.info("Stop process on host %s", host)

    LOG.debug("Send SIGTERM to host %s", host)
    process.send_signal(signal.SIGTERM)
    time.sleep(1)

    if not process.is_running():
        return

    for _ in range(GENTLE_STOP_TIMEOUT - 1):
        if not process.is_running():
            time.sleep(1)
    else:
        LOG.debug("Send SIGKILL to host %s", host)
        process.send_signal(signal.SIGKILL)



# -*- coding: utf-8 -*-


import paramiko
import spur
import spur.ssh


class SSH(spur.SshShell):

    def _connect_ssh(self):
        if self._client is None:
            if self._closed:
                raise RuntimeError("Shell is closed")
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(spur.ssh.MissingHostKey.accept)
            client.connect(
                hostname=self._hostname,
                port=self._port,
                username=self._username,
                password=self._password,
                key_filename=self._private_key_file,
                look_for_keys=self._look_for_private_keys,
                timeout=self._connect_timeout
            )

            # Patch transport for faster SSH
            # https://github.com/paramiko/paramiko/issues/175
            transport = client.get_transport()
            transport.window_size = 2147483647
            transport.packetizer.REKEY_BYTES = pow(2, 40)
            transport.packetizer.REKEY_PACKETS = pow(2, 40)

            self._client = client
        return self._client


def get_ssh(hostname):
    return SSH(hostname, shell_type=spur.ssh.ShellTypes.sh)



# -*- coding: utf-8 -*-


import glob
import os
import os.path
import posixpath

import concurrent.futures

import fuelpdsh
import fuelpdsh.ssh
import fuelpdsh.utils


LOG = fuelpdsh.logger(__name__)
"""Logger."""


BUFFER_LENGTH = 100 * 1024  # kilobytes
"""How long copy buffer has to be."""


def command(hostnames, options, stop_ev):
    files_to_copy = get_local_paths(options.src_path)
    files_to_copy = get_only_files(files_to_copy)

    if not files_to_copy:
        return os.EX_OK

    concurrency = options.concurrency
    if not concurrency:
        concurrency = len(hostnames)

    LOG.info("Execute with %d threads", concurrency)

    futures = []
    copy_pool = concurrent.futures.ThreadPoolExecutor(concurrency * len(files_to_copy))

    try:
        with concurrent.futures.ThreadPoolExecutor(concurrency) as pool:
            for host in hostnames:
                future = pool.submit(cp_to_remote, host, options, files_to_copy, copy_pool, stop_ev)
                futures.append(future)

            fuelpdsh.utils.wait_for_futures(futures, stop_ev)
    finally:
        copy_pool.shutdown()
        return fuelpdsh.utils.futures_exit_code(futures)


def cp_to_remote(host, options, files_to_copy, pool, stop_ev):
    if stop_ev.is_set():
        return os.EX_OK

    LOG.info("Copy %d files to host %s", len(files_to_copy), host)

    futures = []
    with fuelpdsh.ssh.get_ssh(host) as ssh:
        for filename in files_to_copy:
            local_basename = os.path.basename(filename)
            remote_path = posixpath.join(options.dst_path, local_basename)
            future = pool.submit(copy_file, ssh, host, filename, remote_path, stop_ev)
            futures.append(future)

        fuelpdsh.utils.wait_for_futures(futures, stop_ev)

    return fuelpdsh.utils.futures_exit_code(futures)


def copy_file(ssh, host, local_path, remote_path, stop_ev):
    LOG.info("Copy %s to %s:%s", local_path, host, remote_path)

    with ssh.open(remote_path, "wb") as remote_fileobj:
        with open(local_path, "rb") as local_fileobj:
            copy_finished = False

            while not stop_ev.is_set():
                buf = local_fileobj.read(BUFFER_LENGTH)
                if not buf:
                    copy_finished = True
                    break
                remote_fileobj.write(buf)

            if not copy_finished:
                ssh.run(["rm", "-f", remote_path], allow_error=True)
                LOG.warning("File %s was not copied to %s:%s", local_path, host, remote_path)
                return os.EX_SOFTWARE

    if copy_finished:
        LOG.info("File %s was copied to %s:%s", local_path, host, remote_path)

    return os.EX_OK


def get_local_paths(paths):
    paths = [os.path.expanduser(path) for path in paths]

    expanded_paths = []
    already_met_paths = set()
    for path in paths:
        if os.path.isdir(path):
            extracted = expand_directory(path)
        else:
            extracted = glob.glob(path)
        extracted = [os.path.realpath(path_) for path_ in extracted]
        for extracted_path in extracted:
            if extracted_path not in already_met_paths:
                already_met_paths.add(extracted_path)
                expanded_paths.append(extracted_path)

    return expanded_paths


def get_only_files(paths):
    files = []

    for path in paths:
        if not os.path.isfile(path):
            LOG.debug("Skip path %s because it is not a file", path)
            continue
        if not os.access(path, os.R_OK):
            LOG.warning("Skip file %s because it is not readable", path)
            continue
        files.append(path)

    return files


def expand_directory(directory):
    return [os.path.join(directory, path) for path in os.listdir(directory)]



# -*- coding: utf-8 -*-


import codecs
import logging
import Queue
import threading
import time

import fuelpdsh


LOG = fuelpdsh.logger(__name__)
"""Logger."""


class Stream(object):

    QUEUE_LENGTH = 10000
    NO_ITEM = object()

    def __init__(self, original_stream):
        self.stream = original_stream
        self.stop_ev = threading.Event()
        self.queue = Queue.Queue(self.QUEUE_LENGTH)
        self.thread = None

    def run(self):
        self.thread = threading.Thread(target=self.background_task)
        self.thread.daemon = True
        self.thread.start()

        LOG.debug("Background thread %s was started", self.thread.ident)

    def stop(self):
        self.stop_ev.set()

    def wait(self):
        if self.thread is None:
            return

        thread, self.thread = self.thread, None

        LOG.debug("Starting queue flush for thread %s", thread.ident)
        self.queue.join()

        LOG.debug("Waiting for background thread %s to be finished", thread.ident)
        thread.join()

    def background_task(self):
        while not self.stop_ev.is_set():
            item = self.fetch_item()
            if item is self.NO_ITEM:
                time.sleep(0.01)
                continue
            self.process_item(item)

        LOG.debug("Stop event is set for thread %s, start to flush queue", threading.current_thread().ident)

        while not self.queue.empty():
            item = self.fetch_item()
            if item is self.NO_ITEM:
                break
            self.process_item(item)

        LOG.debug("Thread %s finished", threading.current_thread().ident)

    def fetch_item(self):
        try:
            return self.queue.get_nowait()
        except Queue.Empty:
            return self.NO_ITEM

    def put_item(self, data):
        self.queue.put(data, True)

    def process_item(self, item):
        self.stream.write(item)
        self.stream.write("\n")
        self.queue.task_done()

    def make_host_stream(self, hostname, hostname_length):
        return HostStream(self, hostname, hostname_length)


class HostStream(object):

    __slots__ = "prefix", "stream", "accumulator", "closed", "lock"

    DECODER = codecs.getdecoder("unicode_escape")

    def __init__(self, stream, hostname, hostname_length):
        self.prefix = hostname.ljust(hostname_length) + ":  "
        self.stream = stream
        self.accumulator = ""
        self.closed = False
        self.lock = threading.RLock()

    def put(self, data):
        if not self.closed:
            data = self.DECODER(data, "replace")
            self.stream.put_item(self.prefix + data[0])

    def write(self, data):
        chunks = []

        with self.lock:
            self.accumulator += data
            if "\n" in self.accumulator:
                chunks = self.accumulator.rsplit("\n")
                self.accumulator = chunks.pop()

        for chunk in chunks:
            self.put(chunk)

    def close(self):
        with self.lock:
            if self.accumulator:
                self.put(self.accumulator)
                self.accumulator = ""
            self.closed = True

    flush = close



# -*- coding: utf-8 -*-


import argparse
import atexit
import functools
import logging
import os
import re
import threading

import fuelpdsh
import fuelpdsh.pdsh
import fuelpdsh.cp
import fuelpdsh.utils


LOG = fuelpdsh.logger(__name__)
"""Logger."""

DEFAULT_CONCURRENCY = 4
"""How many hosts we have to access in parallel."""


def cli(cmd_to_execute, argumenter):
    stop_ev = threading.Event()
    atexit.register(stop_ev.set)

    options = get_options(argumenter)
    configure(options)

    try:
        nodes = fuelpdsh.utils.get_nodes(options)
    except:
        return os.EX_SOFTWARE

    try:
        return cmd_to_execute(nodes, options, stop_ev)
    except:
        return os.EX_SOFTWARE


def remote_cmd_argumenter(parser):
    parser.add_argument(
        "command",
        help="Command to execute",
        nargs="+"
    )


def cp_to_remote_argumenter(parser):
    parser.add_argument(
        "src_path",
        help="Source paths to copy",
        nargs="+")
    parser.add_argument(
        "dst_path",
        help="Remote destination")


def configure(options):
    level = logging.ERROR
    log_format = "%(message)s"

    if options.verbose:
        level = logging.INFO
        log_format = "*** %(thread)d >>> %(message)s"
    elif options.debug:
        level = logging.DEBUG
        log_format = "%(thread)d | [%(levelname)-5s] (%(module)10s:%(lineno)-3d) %(asctime)-15s: %(message)s"

    for namespace in "fuelpdsh", "paramiko":
        configure_logger(namespace, level, log_format)

    LOG.debug("Options are %s", options)


def configure_logger(namespace, log_level, log_format):
    root_logger = logging.getLogger("fuelpdsh")
    root_logger.handlers = []
    root_logger.propagate = False
    root_logger.setLevel(log_level)

    handler = logging.StreamHandler()
    formatter = logging.Formatter(log_format)
    handler.setFormatter(formatter)
    root_logger.addHandler(handler)


def get_options(argumenter):
    parser = argparse.ArgumentParser(
        epilog="Please contact Sergey Arkhipov <serge@aerialsounds.org> for issues."
    )

    parser.add_argument(
        "--concurrency",
        help="How many simultaneous connections should be established. "
             "By default (%(default)d), we are trying to connect to all nodes, no limits.",
        type=argtype_positive_integer,
        default=DEFAULT_CONCURRENCY
    )
    parser.add_argument(
        "-c", "--cluster-id",
        help="Select only nodes which belong to cluster with such ID.",
        type=int)
    parser.add_argument(
        "-w", "--node-ids",
        help="Plain comma-separated list of nodes.",
        type=argtype_node_ids)
    parser.add_argument(
        "-i", "--ips",
        help="Plain comma-separated list of node IPs.",
        type=argtype_node_ips)
    parser.add_argument(
        "-n", "--name",
        help="Regular expression for the node name.",
        type=re.compile)
    parser.add_argument(
        "-s", "--status",
        help="Node status.")
    parser.add_argument(
        "-g", "--group-id",
        help="Group ID.",
        type=int)
    parser.add_argument(
        "-r", "--roles",
        help="Node roles.",
        type=argtype_roles
    )

    verbosity = parser.add_mutually_exclusive_group(required=False)
    verbosity.add_argument(
        "-v", "--verbose",
        help="Be verbose.",
        action="store_true",
        default=False
    )
    verbosity.add_argument(
        "-d", "--debug",
        help="Be event more verbose, for debugging.",
        action="store_true",
        default=False
    )

    argumenter(parser)

    return parser.parse_args()


def argtype_comma_separated_list(func):
    @functools.wraps(func)
    def decorator(value):
        if not isinstance(value, basestring):
            raise argparse.ArgumentTypeError("Value {0} has to be a string".format(value))
        if not value:
            value = []
        else:
            value = value.rstrip(",").split(",")

        return func(value)

    return decorator


@argtype_comma_separated_list
def argtype_roles(values):
    return values


@argtype_comma_separated_list
def argtype_node_ids(values):
    converted_values = []

    for value in values:
        if not value.startswith("node-"):
            value = "node-" + value
        converted_values.append(value)

    return converted_values


@argtype_comma_separated_list
def argtype_node_ips(values):
    failed_values = []

    ip_regex = re.compile(
        r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)")
    for value in values:
        if not ip_regex.match(value):
            failed_values.append(value)

    if failed_values:
        raise argparse.ArgumentTypeError(
            "Values {0} are not correct IPs".format(", ".join(failed_values)))

    return values


def argtype_positive_integer(value):
    try:
        value = int(value)
    except ValueError:
        pass
    else:
        if value >= 0:
            return value

    raise argparse.ArgumentTypeError("Value {0} has to be a positive integer".format(value))


remote_cmd = functools.partial(cli, fuelpdsh.pdsh.command, remote_cmd_argumenter)
cp_to_remote = functools.partial(cli, fuelpdsh.cp.command, cp_to_remote_argumenter)



# -*- coding: utf-8 -*-


import os
import time

import concurrent.futures
import fuelclient

import fuelpdsh


LOG = fuelpdsh.logger(__name__)
"""Logger."""


def wait_for_futures(futures, stop_ev):
    try:
        while not all(future.done() for future in futures):
            time.sleep(1)
    except KeyboardInterrupt:
        LOG.debug("Set stop event")
    finally:
        stop_ev.set()
        while not all(future.done() for future in futures):
            time.sleep(0.01)


def futures_exit_code(futures):
    exit_code = os.EX_OK

    for future in futures:
        try:
            result = future.result()
        except concurrent.futures.CancelledError:
            result = os.EX_SOFTWARE
        finally:
            if not isinstance(result, int):
                result = os.EX_SOFTWARE
            exit_code = max(exit_code, result)

    return exit_code


def get_nodes(options):
    client = fuelclient.get_client("node")

    conditions = []
    if options.node_ids:
        LOG.debug("Filter on node IDs %s", options.node_ids)
        conditions.append(lambda node: node["hostname"] in options.node_ids)
    if options.ips:
        LOG.debug("Filter on node IPs %s", options.ips)
        conditions.append(lambda node: node["ip"] in options.ips)
    if options.name:
        LOG.debug("Filter on node regexp name '%s'", options.name.pattern)
        conditions.append(lambda node: options.name.search(node["name"]))
    if options.status:
        LOG.debug("Filter on node status %s", options.status)
        conditions.append(lambda node: node["status"] == options.status)
    if options.roles:
        LOG.debug("Filter on node roles %s", options.roles)
        conditions.append(lambda node: set(node["roles"]) & set(options.roles))
    if options.group_id:
        LOG.debug("Filter on node group ID %s", options.group_id)
        conditions.append(lambda node: node["group_id"] == options.group_id)

    try:
        nodes = client.get_all(options.cluster_id)
    except Exception as exc:
        LOG.error("Cannot fetch from Fuel: %s", exc)
        raise Exception

    nodes = sorted(node["hostname"] for node in nodes if all(cond(node) for cond in conditions))

    LOG.info("Found %d suitable nodes", len(nodes))
    LOG.debug("Nodes to execute on: %s", nodes)

    return nodes



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import setuptools

try:
    import multiprocessing
    assert multiprocessing
except ImportError:
    pass


setuptools.setup(
    setup_requires=["pbr"],
    pbr=True)
